/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpi project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpi
 *
 *  Created on: Feb 14, 2011
 *      Author: Martin Schreiber (martin.schreiber@in.tum.de)
 */

#ifndef TRAVERSATORS_CADAPTIVE_CONFORMING_GRID_VERTICES_DEPTHLIMITER_HPP_
#define TRAVERSATORS_CADAPTIVE_CONFORMING_GRID_VERTICES_DEPTHLIMITER_HPP_

#include <assert.h>
#include "config.h"
#include "libsierpi/stacks/CSimulationStacks.hpp"
#include "libsierpi/triangle/CTriangle_Factory.hpp"
#include "libsierpi/triangle/CTriangle_Normals.hpp"
#include "libsierpi/parallelization/CPartition_SplitJoinInformation.hpp"
#include "CAutomatonTable.hpp"

namespace sierpi
{
namespace travs
{

/*******************************************************
 * !!!!!!!!!! FIRST TRAVERSAL !!!!!!!!!!!
 *******************************************************/

template <typename CKernelClass, typename CSimulationStacks>
class CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_FirstTraversal
{
	typedef typename CSimulationStacks::TElementData TElementData;

public:
	CKernelClass &kernelClass;

	/**
	 * a few typedefs which could be utilized by the traversator to get more informations
	 */
	typedef typename CKernelClass::TVertexScalar TVertexScalar;
	typedef CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_FirstTraversal<CKernelClass,CSimulationStacks> TThisClass;

	/**
	 * stack and lifo handlers used in this traversator
	 */
	CStack<char> *structure_stack_out;
	CStackReaderTopDown<char> structure_lifo_in;

	/**
	 * lifo iterator and output stack for element data
	 */
	CStackReaderTopDown<TElementData> element_lifo_in;

	/**
	 * stack to store element wise refine/coarsen automata states
	 */
	CStack<unsigned char> *adaptivity_state_flag_stack_out;

public:
	/**
	 * pointer to the corresponding communication stacks
	 */
	CStack<char> *comm_left_edge_stack;
	CStack<char> *comm_right_edge_stack;

	/**
	 * repeat_traversal is set to true, if another refinement iteration has to
	 * be done in order to achieve a conforming grid
	 */
	bool repeat_traversal;


#include "auto/CAdaptiveConformingGrid_Vertices_DepthLimiters_Element_FirstTraversal_autogenerated.hpp"

/**
 * helper functions usage by every parallel traversator
 */
#include "../CTraversatorClassInc_SetupAdaptiveFirstPass.hpp"


	void setup_TraversatorParameters(
			int p_depth_limiter_min,
			int p_depth_limiter_max
	)
	{
		depth_limiter_min = p_depth_limiter_min;
		depth_limiter_max = p_depth_limiter_max;
	}

	CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_FirstTraversal(CKernelClass &p_kernelClass)	:
				kernelClass(p_kernelClass)
	{
		depth_limiter_min = -1;
		depth_limiter_max = -1;
	}

	bool action(CSimulationStacks *cStacks)
	{
		assert(depth_limiter_min != -1);
		assert(depth_limiter_max != -1);

		// INPUT structure stack
		structure_lifo_in.setup(cStacks->structure_stacks.forward);

		// OUTPUT structure stack to output stack in reversed order
		structure_stack_out = &cStacks->structure_stacks.backward;

		element_lifo_in.setup(cStacks->element_data_stacks.forward);

		comm_left_edge_stack = &cStacks->adaptive_comm_left_edge_stack;
		comm_right_edge_stack = &cStacks->adaptive_comm_right_edge_stack;

		adaptivity_state_flag_stack_out = &(cStacks->adaptivity_state_flag_stacks.backward);

		repeat_traversal = false;

		kernelClass.traversal_pre_hook();

		(this->*sfcRecursiveMethod_Forward)(
				triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
				triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
				triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
				triangleFactory.recursionDepthFirstRecMethod
		);

		if (cStacks->structure_stacks.forward.getNumberOfElementsOnStack() == 1)
		{
			/*
			 * check for coarsening request and change it to 'nothing to do'
			 *
			 * the information is encoded in the lower 4 bits
			 */
			if ((adaptivity_state_flag_stack_out->getElementAtIndex(0)&7) == 2)
			{

				adaptivity_state_flag_stack_out->setElementAtIndex(0, 0);
			}
		}

		kernelClass.traversal_post_hook();

		assert(structure_stack_out->structure_isValidTriangle());
		assert(structure_stack_out->structure_getNumberOfTrianglesInTriangle() == adaptivity_state_flag_stack_out->getNumberOfElementsOnStack());

		if (repeat_traversal)
			return true;

		// 0 => nothing
		// 1 => refine
		// 2 => coarsen
		if (comm_left_edge_stack->query_anyElementNotEqualTo_Scalar(0))
			return true;
		if (comm_right_edge_stack->query_anyElementNotEqualTo_Scalar(0))
			return true;

		return false;
	}
};



/*******************************************************
 * !!!!!!!!!! MIDDLE TRAVERSAL !!!!!!!!!!!
 *******************************************************/

template <typename CKernelClass, typename CSimulationStacks>
class CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_MiddleTraversals
{
public:
	typedef typename CSimulationStacks::TElementData TElementData;

	CKernelClass &kernelClass;

	/**
	 * a few typedefs which could be utilized by the traversator to get more informations
	 */
	typedef typename CKernelClass::TVertexScalar TVertexScalar;
	typedef CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_MiddleTraversals<CKernelClass,CSimulationStacks> TThisClass;

	/**
	 * stack and lifo handlers used in this traversator
	 */
	CStackReaderTopDown<char> structure_lifo_in;

	/**
	 * stack to store element wise refine/coarsen automata states
	 */
	CStack<unsigned char> *adaptivity_state_flag_stack_in;
	CStack<unsigned char> *adaptivity_state_flag_stack_out;

	/**
	 * pointer to the corresponding communication stacks
	 */
	CStack<char> *comm_left_edge_stack;
	CStack<char> *comm_right_edge_stack;


	/**
	 * repeat_traversal is set to true, if another refinement iteration has to
	 * be done in order to achieve a conforming grid
	 */
	bool repeat_traversal;

#include "auto/CAdaptiveConformingGrid_Vertices_DepthLimiters_Element_MiddleTraversal_autogenerated.hpp"

	/**
	 * helper functions usage by every parallel traversator
	 */
#include "../CTraversatorClassInc_SetupAdaptiveMiddlePasses.hpp"

	CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_MiddleTraversals(CKernelClass &p_kernelClass)	:
		kernelClass(p_kernelClass)
	{
	}


	/**************************************************************************
	 * backward and forward traversals
	 *
	 * \return	true is returned when this action has to be repeated
	 **************************************************************************/
	bool action_Serial(CSimulationStacks *cStacks)
	{
		structure_lifo_in.setup(cStacks->structure_stacks.backward);

		comm_left_edge_stack = &cStacks->adaptive_comm_left_edge_stack;
		comm_right_edge_stack = &cStacks->adaptive_comm_right_edge_stack;

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.forward;

		repeat_traversal = false;

		// RECURSIVE FUNCTION CALL
		(this->*sfcRecursiveMethod_Backward)();

		assert(comm_left_edge_stack->isEmpty());
		assert(comm_right_edge_stack->isEmpty());
		assert(adaptivity_state_flag_stack_in->isEmpty());


		/**************************************************************************
		 * forward traversal (all edges adjacent to partitions are set to 'new')
		 **************************************************************************/

		// repeat traversal can also be set to true by the backward traversal due to coarsening reasons.
		// therefore it may not be set to false at this point
		//repeat_traversal = false;

		// use default edge stack as adjacent information communication stack
		comm_left_edge_stack = &cStacks->adaptive_comm_left_edge_stack;
		comm_right_edge_stack = &cStacks->adaptive_comm_right_edge_stack;
		comm_left_edge_stack->clear();
		comm_right_edge_stack->clear();

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.forward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.backward;

		structure_lifo_in.setup(cStacks->structure_stacks.forward);

//		repeat_traversal = false;

		// RECURSIVE FUNCTION CALL
		(this->*sfcRecursiveMethod_Forward)();

		// final checks
		assert(adaptivity_state_flag_stack_in->isEmpty());


		/**************************************************************************
		 * postprocessing
		 **************************************************************************/

		if (repeat_traversal)
			return true;

		/**
		 * check if a value on the edge is equal to 1.
		 * this means, that there's still a refinement to do since this refinement
		 * request is forwarded to an adjacent partition.
		 *
		 * coarsening is local only. this means, that only a close neighborhood of one
		 * triangle may be coarsened. therefore a check for this is not necessary!
		 */

		if (comm_left_edge_stack->query_anyElementEqualTo_Scalar(1))
			return true;
		if (comm_right_edge_stack->query_anyElementEqualTo_Scalar(1))
			return true;

		return false;
	}


	/**************************************************************************
	 * backward and forward traversals
	 *
	 * \return	true is returned when this action has to be repeated
	 **************************************************************************/
	bool action_Parallel(CSimulationStacks *cStacks)
	{
		structure_lifo_in.setup(cStacks->structure_stacks.backward);

		comm_left_edge_stack = &cStacks->adaptive_comm_exchange_right_edge_stack;
		comm_right_edge_stack = &cStacks->adaptive_comm_exchange_left_edge_stack;

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.forward;

		repeat_traversal = false;

		// RECURSIVE FUNCTION CALL
		(this->*sfcRecursiveMethod_Backward)();

		assert(comm_left_edge_stack->isEmpty());
		assert(comm_right_edge_stack->isEmpty());
		assert(adaptivity_state_flag_stack_in->isEmpty());


		/**************************************************************************
		 * forward traversal (all edges adjacent to partitions are set to 'new')
		 **************************************************************************/

		// repeat traversal can also be set to true by the backward traversal due to coarsening reasons.
		// therefore it may not be set to false at this point
		//repeat_traversal = false;

		structure_lifo_in.setup(cStacks->structure_stacks.forward);
		
		// use default edge stack as adjacent information communication stack
		comm_left_edge_stack = &cStacks->adaptive_comm_left_edge_stack;
		comm_right_edge_stack = &cStacks->adaptive_comm_right_edge_stack;
		comm_left_edge_stack->clear();
		comm_right_edge_stack->clear();

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.forward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.backward;

//		repeat_traversal = false;

		// RECURSIVE FUNCTION CALL
		(this->*sfcRecursiveMethod_Forward)();

		// final checks
		assert(adaptivity_state_flag_stack_in->isEmpty());


		/**************************************************************************
		 * postprocessing
		 **************************************************************************/

		if (repeat_traversal)
			return true;

		/**
		 * check if a value on the edge is equal to 1.
		 * this means, that there's still a refinement to do since this refinement
		 * request is forwarded to an adjacent partition.
		 *
		 * coarsening is local only. this means, that only a close neighborhood of one
		 * triangle may be coarsened. therefore a check for this is not necessary!
		 */

		if (comm_left_edge_stack->query_anyElementEqualTo_Scalar(1))
			return true;
		if (comm_right_edge_stack->query_anyElementEqualTo_Scalar(1))
			return true;

		return false;
	}
};



/*************************************************************************
 * !!!!!!!!!! LAST TRAVERSAL !!!!!!!!!!!
 *
 * this traversal also has to implement the split/join information interface
 *************************************************************************/


template <typename CKernelClass, typename CSimulationStacks>
class CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_LastTraversal
{
public:
	typedef typename CSimulationStacks::TElementData TElementData;
	CKernelClass &kernelClass;

	/**
	 * a few typedefs which could be utilized by the traversator to get more informations
	 */
	typedef typename CKernelClass::TVertexScalar TVertexScalar;
	typedef CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_LastTraversal<CKernelClass,CSimulationStacks> TThisClass;

	/**
	 * stack and lifo handlers used in this traversator
	 */
	CStack<char> *structure_stack_in;
	CStack<char> *structure_stack_out;
	CStackReaderTopDown<char> structure_lifo_in;

	/**
	 * Fifo iterator and output stack for element data
	 */
	CStackReaderBottomUp<TElementData> element_fifo_in;
	CStack<TElementData> *element_data_stack_out;

	/**
	 * stack to store element wise refine/coarsen automata states
	 */
	CStack<unsigned char> *adaptivity_state_flag_stack_in;
	CStack<unsigned char> *adaptivity_state_flag_stack_out;

	/**
	 * pointer to the corresponding communication stacks
	 */
	CStack<char> *comm_left_edge_stack;
	CStack<char> *comm_right_edge_stack;

	/**
	 * repeat_traversal is set to true, if another refinement iteration has to
	 * be done in order to achieve a conforming grid
	 */
	bool repeat_traversal;

	/**
	 * maximum depth of the previous traversal
	 */
	int maxLeafTraversalDepth;

#include "auto/CAdaptiveConformingGrid_Vertices_DepthLimiters_Element_LastTraversal_autogenerated.hpp"

/**
 * helper functions usage by every parallel traversator
 */
#include "../CTraversatorClassInc_SetupAdaptiveLastPass.hpp"

	/// information which has to be created for the split/join operations supporting parallelization
	CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_LastTraversal(
			CKernelClass &p_kernelClass
	)	:
			kernelClass(p_kernelClass),
			maxLeafTraversalDepth(-1)
	{
	}



	int action_Parallel(
			CSimulationStacks *cStacks,
			CPartition_SplitJoinInformation &p_splitJoinInformation
	)
	{
#if ADAPTIVE_SUBPARTITION_STACKS
		/*
		 * estimate new size of element data stack.
		 *
		 * this is only an estimation since e. g. coarsening can be forbidden.
		 * therefore we also add an additional offset.
		 *
		 * since the sub-partitions should have cache size, this stack is expected to
		 * be still in cache.
		 */
		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;

		size_t estimated_num_triangles = cStacks->element_data_stacks.forward.getNumberOfElementsOnStack();
		size_t number_of_adaptive_information = adaptivity_state_flag_stack_in->getNumberOfElementsOnStack();

		for (size_t i = 0; i < number_of_adaptive_information; i++)
		{
			unsigned char c = adaptivity_state_flag_stack_in->getElementAtIndex(i) & 7;
			if (c == 2)
			{
				/*
				 * special handling of coarsening
				 */
				estimated_num_triangles--;
				i++;
				continue;
			}

			static int deltaTriangles[8] = {0, 0xfffffff, -1, 1, 1, 2, 2, 3};
			estimated_num_triangles += deltaTriangles[c];
		}

		cStacks->changeStackSize_BeforeAdaptivityLastTraversal(estimated_num_triangles);
#endif

		// input/output structure stack
		structure_stack_in = &cStacks->structure_stacks.backward;
		structure_stack_out = &cStacks->structure_stacks.forward;

		cStacks->adaptive_comm_right_edge_stack.clear();
		cStacks->adaptive_comm_left_edge_stack.clear();

		// left and right are exchanged since the first traversal is backwards!
		comm_left_edge_stack = &cStacks->adaptive_comm_right_edge_stack;
		comm_right_edge_stack = &cStacks->adaptive_comm_left_edge_stack;

		assert(cStacks->element_data_stacks.backward.isEmpty());

		element_fifo_in.setup(cStacks->element_data_stacks.forward);
		element_data_stack_out = &cStacks->element_data_stacks.backward;

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.forward;


		/*
		 * Add offset.
		 *
		 * TODO: Is this really necessary.
		 */
		//estimated_num_triangles += 0;


		// clear output structure stack
		// TODO: is this really necessary?
		structure_stack_out->clear();

		maxLeafTraversalDepth = triangleFactory.recursionDepthFirstRecMethod;

		// RECURSIVE FUNCTIONS

		/***********************************************************
		 * RECURSIVE CALLS - 1ST CHILD
		 ***********************************************************/

		// is this even or odd?
		bool even = (triangleFactory.evenOdd == CTriangle_Enums::EVEN);

		// stack storage place for the hypotenuse
		CStack<char> *hyp_edge_stack = (even ? comm_left_edge_stack : comm_right_edge_stack);

		/***********************************************************
		 * TODO: check if this subtriangle only exists out of a single triangle.
		 * then we have to immediately abort since there's no way to run a split operation on it.
		 * since we can assume to have a basic workload of a few triangles, this shouldn't be a problem
		 */

		/***********************************************************
		 * Sketches given for EVEN triangle factory
		 *
		 * Before split with subtree enumeration
		 *
		 * |\
		 * | 1\
		 * |____\
		 * |    /|\
		 * | 2/3 | 4\
		 * |/____|____\
		 *
		 *
		 * after split
		 * |\
		 * | 1\
		 * |____\
		 * |    /
		 * | 2/   /|\
		 * |/   /3 | 4\ left edge comm stack (identified after traversal of subtree 1) [backward]
		 *    /____|____\
		 *      right edge comm stack [backward]
		 *
		 *
		 * REMEMBER: THE ACTUAL TRAVERSAL IS BACKWARDS!!!
		 */

		TVertexScalar mx = (triangleFactory.vertices[0][0] + triangleFactory.vertices[1][0])*(TVertexScalar)0.5;
		TVertexScalar my = (triangleFactory.vertices[0][1] + triangleFactory.vertices[1][1])*(TVertexScalar)0.5;
#if 0
		if (triangleFactory.recursionDepthFirstRecMethod >= triangleFactory.minDepth)
		{
			/**************************************
			 * max splitting depth reached
			 **************************************/
			p_splitJoinInformation.splitting_permitted = false;

#ifdef DEBUG
			int retval =
#endif
					(this->*recursionTable[recursionMethodId].parent_method)(
					triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
					triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
					triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
					triangleFactory.recursionDepthFirstRecMethod
				);

#ifdef DEBUG
			assert(retval == 1);
#endif
		}
		else
#endif
		if (structure_stack_in->pop() == 0) // pop root level structure stack element
		{
			/**************************************
			 * there's only a single triangle
			 **************************************/
			p_splitJoinInformation.splitting_permitted = false;

			structure_stack_in->push(0);

			int retval =
			(this->*sfcRecursiveMethod_Backward)(
					triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
					triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
					triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
					triangleFactory.recursionDepthFirstRecMethod
				);

			if (retval == 0)
			{
				/*
				 * coarsening requested => ignore and fix invalid stacks
				 *
				 * TODO: do we have to fix the comm stacks?
				 * maybe this stacks are not exchanged with adjacent subpartitions
				 */
				for (size_t i = 0; i < comm_left_edge_stack->getNumberOfElementsOnStack(); i++)
					comm_left_edge_stack->setElementAtIndex(i, 0);

				for (size_t i = 0; i < comm_right_edge_stack->getNumberOfElementsOnStack(); i++)
					comm_right_edge_stack->setElementAtIndex(i, 0);

				/*
				 * since the traverser did not store anything on the structure stack, assuming that it is
				 * allowed to coarsen, we have to push an '0' to the structure stack
				 */
				structure_stack_out->push(0);
				element_data_stack_out->push_PtrValue(element_fifo_in.getPreviousDataPtr());
			}

			p_splitJoinInformation.splitting_permitted = false;

			/**
			 * create split/join information for single triangle
			 */

			/**
			 * setup data for pseudo first sub-triangle, but not for second one!
			 */
			p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_right_stack =
					comm_left_edge_stack->getNumberOfElementsOnStack();
			p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_left_stack =
					comm_right_edge_stack->getNumberOfElementsOnStack();
			p_splitJoinInformation.first_triangle.number_of_elements = 1;

			p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack = 0;
			p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack = 0;
			p_splitJoinInformation.second_triangle.number_of_elements = 0;

			p_splitJoinInformation.number_of_shared_edge_comm_elements = 0;
		}
		else
		{
			/**************************************
			 * there are at least 2 subtriangles for this subpartition!
			 ***************************************/

			/**************************************
			 * a split operation is not allowed whenever there are only 2 triangles and both request a coarsening
			 **************************************/
#if 0
			/**************************************
			 * a simple solution is to forbid splittings below the minimum recursion depth
			 **************************************/
			p_splitJoinInformation.splitting_permitted = (triangleFactory.recursionDepthFirstRecMethod < triangleFactory.minDepth);
#else
			/**************************************
			 * a more elegant solution is to set the permission to false whenever a coarsening step is taken after this step
			 **************************************/
			p_splitJoinInformation.splitting_permitted = true;
#endif

			/*
			 * backup start stack index to get shared edges
			 */
			int shared_edges_start_counter;

			// pop 2nd level structure stack element
			if (structure_stack_in->pop() == 0)
			{
				/*****************************************************
				 * FIRST sub-subtriangle is also only a single triangle without further refinements
				 *****************************************************/

				// this sub-triangle only has only a recursion level of 1
				maxLeafTraversalDepth = triangleFactory.recursionDepthFirstRecMethod+1;

				// no more leaves => single triangle for the first subtriangle
				// push back the 0 value
				structure_stack_in->push(0);

				if (triangleFactory.evenOdd == CTriangle_Enums::EVEN)
				{
					(this->*sfcRecursiveMethod_Backward_1stChild)(
							triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							mx, my,
							triangleFactory.recursionDepthFirstRecMethod+1
					);
				}
				else
				{
					(this->*sfcRecursiveMethod_Backward_1stChild)(
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
							mx, my,
							triangleFactory.recursionDepthFirstRecMethod+1
					);
				}

				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack =
						comm_left_edge_stack->getNumberOfElementsOnStack();

				// remember edge comm data on the right communication stack
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack =
						comm_right_edge_stack->getNumberOfElementsOnStack();

				p_splitJoinInformation.second_triangle.number_of_elements = element_data_stack_out->getNumberOfElementsOnStack();

				p_splitJoinInformation.number_of_shared_edge_comm_elements = 1;

				shared_edges_start_counter = hyp_edge_stack->getNumberOfElementsOnStack()-p_splitJoinInformation.number_of_shared_edge_comm_elements;
			}
			else
			{
				/****************************************************
				 * FIRST subtriangle can be further traversed down
				 *****************************************************/
				int retval2;
				if (triangleFactory.evenOdd == CTriangle_Enums::EVEN)
				{
					TVertexScalar nmx = (triangleFactory.vertices[1][0] + triangleFactory.vertices[2][0])*0.5;
					TVertexScalar nmy = (triangleFactory.vertices[1][1] + triangleFactory.vertices[2][1])*0.5;

					(this->*sfcRecursiveMethod_Backward_1stChild_1stChild)
						(
							mx, my,
							triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
							nmx, nmy,
							triangleFactory.recursionDepthFirstRecMethod+2
						);

					shared_edges_start_counter = hyp_edge_stack->getNumberOfElementsOnStack();

					retval2 = (this->*sfcRecursiveMethod_Backward_1stChild_2ndChild)
						(
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							mx, my,
							nmx, nmy,
							triangleFactory.recursionDepthFirstRecMethod+2
						);
				}
				else
				{
					TVertexScalar nmx = (triangleFactory.vertices[0][0] + triangleFactory.vertices[2][0])*0.5;
					TVertexScalar nmy = (triangleFactory.vertices[0][1] + triangleFactory.vertices[2][1])*0.5;

					(this->*sfcRecursiveMethod_Backward_1stChild_1stChild)
						(
							triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
							mx, my,
							nmx, nmy,
							triangleFactory.recursionDepthFirstRecMethod+2
						);

					shared_edges_start_counter = hyp_edge_stack->getNumberOfElementsOnStack();

					retval2 = (this->*sfcRecursiveMethod_Backward_1stChild_2ndChild)
						(
							mx, my,
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							nmx, nmy,
							triangleFactory.recursionDepthFirstRecMethod+2
						);
				}

				// RECURSIVE CALLS for SUBTRIANGLE_2

				/*
				 * when retval2 == 1, no further operation has to be done.
				 * if retval2 == 0, both triangles should be joined.
				 */
				if (retval2 == 1)
				{
					structure_stack_out->push(1);
				}
				else
				{
					// after coarsening, this sub-triangle only has only a recursion level of 1
					maxLeafTraversalDepth = triangleFactory.recursionDepthFirstRecMethod+1;

					/*
					 * since retval2 == 0, both subtriangles like to be joined
					 */
					structure_stack_out->push(0);
					TElementData *element = element_data_stack_out->fakePush_ReturnPtr();

					TElementData *right_element;
					TElementData *left_element;
					if (triangleFactory.evenOdd == CTriangle_Enums::EVEN)
					{
						right_element = element_fifo_in.getPreviousDataPtr();
						left_element = element_fifo_in.getPreviousPreviousDataPtr();
					}
					else
					{
						right_element = element_fifo_in.getPreviousPreviousDataPtr();
						left_element = element_fifo_in.getPreviousDataPtr();
					}

					if (triangleFactory.evenOdd == CTriangle_Enums::EVEN)
					{
						kernelClass.coarsen(
								triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
								triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
								mx, my,
								left_element, right_element, element
							);
					}
					else
					{
						kernelClass.coarsen(
								triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
								triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
								mx, my,
								left_element, right_element, element
							);
					}
				}

				/***************************************************************************************
				 * BACKUP information about SECOND TRIANGLE:
				 *
				 * remember edge comm data on the left communication stack.
				 * the number has to be stored for the right stack since this traversal goes backwards.
				 */
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack =
						comm_left_edge_stack->getNumberOfElementsOnStack();

				// remember edge comm data on the right communication stack
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack =
						comm_right_edge_stack->getNumberOfElementsOnStack();

				// remember number of elements on the element data stack
				p_splitJoinInformation.second_triangle.number_of_elements = element_data_stack_out->getNumberOfElementsOnStack();


				/***************************************************************************************
				 * BACKUP adaptive refinement/coarsening information on splitting edge
				 *
				 * we have to search the 'number_of_shared_edge_comm_elements' topmost
				 * elements on the left stack for occurrences of refinement or coarsening informations
				 */
				p_splitJoinInformation.number_of_shared_edge_comm_elements = hyp_edge_stack->getNumberOfElementsOnStack() - shared_edges_start_counter;
			}

			bool within_coarsening = false;
			int number_of_shared_edge_comm_elements_delta_elements = 0;

			assert(p_splitJoinInformation.number_of_shared_edge_comm_elements == ((int)hyp_edge_stack->getNumberOfElementsOnStack() - shared_edges_start_counter));

			for (int i = hyp_edge_stack->getNumberOfElementsOnStack()-1; i >= shared_edges_start_counter; i--)
			{
				int a = hyp_edge_stack->getElementAtIndex(i);

				if (a == 0)
					continue;

				if (a == 1)
				{
					number_of_shared_edge_comm_elements_delta_elements++;
					continue;
				}

				assert(a == -1);

				within_coarsening = !within_coarsening;

				if (within_coarsening)
					number_of_shared_edge_comm_elements_delta_elements--;
			}
			assert(within_coarsening == false || p_splitJoinInformation.number_of_shared_edge_comm_elements == 1);



			/***************************************************************************************
			 * RECURSIVE CALLS - 2ND CHILD
			 */
			// pop 2nd level structure stack element

			int retval;
			if (triangleFactory.evenOdd == CTriangle_Enums::EVEN)
			{
				retval = (this->*sfcRecursiveMethod_Backward_2ndChild)(
										triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
										triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
										mx, my,
										triangleFactory.recursionDepthFirstRecMethod+1
									);
			}
			else
			{
				retval = (this->*sfcRecursiveMethod_Backward_2ndChild)(
										triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
										triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
										mx, my,
										triangleFactory.recursionDepthFirstRecMethod+1
									);
			}


			// RECURSIVE CALLS for SUBTRIANGLE_2
			if (retval != 0)
			{
				structure_stack_out->push(retval);
			}
			else
			{
				// coarsen
				structure_stack_out->push(0);
				TElementData *element = element_data_stack_out->fakePush_ReturnPtr();


				TElementData *right_element;
				TElementData *left_element;

				if (triangleFactory.evenOdd == CTriangle_Enums::ODD)
				{
					right_element = element_fifo_in.getPreviousDataPtr();
					left_element = element_fifo_in.getPreviousPreviousDataPtr();
				}
				else
				{
					right_element = element_fifo_in.getPreviousPreviousDataPtr();
					left_element = element_fifo_in.getPreviousDataPtr();
				}


				if (triangleFactory.evenOdd == CTriangle_Enums::EVEN)
				{
					kernelClass.coarsen(
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
							mx, my,
							left_element, right_element, element
						);
				}
				else
				{
					kernelClass.coarsen(
							triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
							triangleFactory.vertices[2][0], triangleFactory.vertices[2][1],
							mx, my,
							left_element, right_element, element
						);
				}

				p_splitJoinInformation.splitting_permitted = false;
			}

			p_splitJoinInformation.first_triangle.number_of_elements =
					element_data_stack_out->getNumberOfElementsOnStack()
					- p_splitJoinInformation.second_triangle.number_of_elements;

			if (even)
			{
				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_left_stack =
						comm_right_edge_stack->getNumberOfElementsOnStack()
						- p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack;

				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_right_stack =
						comm_left_edge_stack->getNumberOfElementsOnStack()
						- p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack
						+ 2*p_splitJoinInformation.number_of_shared_edge_comm_elements;

				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_right_stack += number_of_shared_edge_comm_elements_delta_elements;
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack += number_of_shared_edge_comm_elements_delta_elements;
			}
			else
			{
				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_right_stack =
						comm_left_edge_stack->getNumberOfElementsOnStack()
						- p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_right_stack;

				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_left_stack =
						comm_right_edge_stack->getNumberOfElementsOnStack()
						- p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack
						+ 2*p_splitJoinInformation.number_of_shared_edge_comm_elements;
						// 2 times, since the shared edge comm elements are already included in number_of_edge_comm_elements_left_stack

				p_splitJoinInformation.first_triangle.number_of_edge_comm_elements_left_stack += number_of_shared_edge_comm_elements_delta_elements;
				p_splitJoinInformation.second_triangle.number_of_edge_comm_elements_left_stack += number_of_shared_edge_comm_elements_delta_elements;
			}


			/***********************************************************
			 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			 * update new shared edge comm elements right now.
			 * !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			 */
			p_splitJoinInformation.number_of_shared_edge_comm_elements += number_of_shared_edge_comm_elements_delta_elements;
		} // processing of at least 2 triangles in this sub-partition

		// debug
		assert(structure_stack_out->structure_isValidTriangle());
		assert(structure_stack_out->structure_getNumberOfTrianglesInTriangle() == element_data_stack_out->getNumberOfElementsOnStack());

		// clear forward stack and SWAP to make the output stack the forward stack
		cStacks->element_data_stacks.forward.clear();
		cStacks->element_data_stacks.swap();

		kernelClass.traversal_post_hook();

		p_splitJoinInformation.joining_permitted = true;

#if ADAPTIVE_SUBPARTITION_STACKS
		/*
		 * if this assertion fails, add some offset to estimated_num_triangles
		 */
		assert (estimated_num_triangles == cStacks->element_data_stacks.forward.getNumberOfElementsOnStack());

		cStacks->changeStackSize_AfterAdaptivityLastTraversal();
#endif
		return maxLeafTraversalDepth;
	}


	/**
	 * action without updating the split/join information
	 */
	int action_Serial(CSimulationStacks *cStacks)
	{
#if ADAPTIVE_SUBPARTITION_STACKS
		/*
		 * estimate new size of element data stack.
		 *
		 * this is only an estimation since e. g. coarsening can be forbidden.
		 * therefore we also add an additional offset.
		 *
		 * since the sub-partitions should have cache size, this stack is expected to
		 * be still in cache.
		 */
		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;

		size_t estimated_num_triangles = cStacks->element_data_stacks.forward.getNumberOfElementsOnStack();
		size_t number_of_adaptive_information = adaptivity_state_flag_stack_in->getNumberOfElementsOnStack();

		for (size_t i = 0; i < number_of_adaptive_information; i++)
		{
			unsigned char c = adaptivity_state_flag_stack_in->getElementAtIndex(i) & 7;
			if (c == 2)
			{
				/*
				 * special handling of coarsening
				 */
				estimated_num_triangles--;
				i++;
				continue;
			}

			static int deltaTriangles[8] = {0, 0xfffffff, -1, 1, 1, 2, 2, 3};
			estimated_num_triangles += deltaTriangles[c];
		}

		cStacks->changeStackSize_BeforeAdaptivityLastTraversal(estimated_num_triangles);
#endif
		/**
		 * SETUP LAST TRAVERSAL
		 */

		cStacks->adaptive_comm_right_edge_stack.clear();
		cStacks->adaptive_comm_left_edge_stack.clear();

		element_data_stack_out = &cStacks->element_data_stacks.backward;

		assert(element_data_stack_out->isEmpty());

		adaptivity_state_flag_stack_in = &cStacks->adaptivity_state_flag_stacks.backward;
		adaptivity_state_flag_stack_out = &cStacks->adaptivity_state_flag_stacks.forward;

		element_fifo_in.setup(cStacks->element_data_stacks.forward);

		// store the output to the backward stack (backward is swapped with the forward stack later on)
		// input/output structure stack
		structure_stack_in = &cStacks->structure_stacks.backward;
		structure_stack_out = &cStacks->structure_stacks.forward;

		// left and right are exchanged since the first traversal is backwards!
		comm_left_edge_stack = &cStacks->adaptive_comm_right_edge_stack;
		comm_right_edge_stack = &cStacks->adaptive_comm_left_edge_stack;

		// clear output structure stack
		// TODO: is this really necessary?
		structure_stack_out->clear();

		maxLeafTraversalDepth = 0;

		kernelClass.traversal_pre_hook();

		(this->*sfcRecursiveMethod_Backward)(
				triangleFactory.recursionDepthFirstRecMethod,
				triangleFactory.vertices[0][0], triangleFactory.vertices[0][1],
				triangleFactory.vertices[1][0], triangleFactory.vertices[1][1],
				triangleFactory.vertices[2][0], triangleFactory.vertices[2][1]
				);

		kernelClass.traversal_post_hook();

		// debug
		assert(structure_stack_out->structure_isValidTriangle());
		assert(structure_stack_out->structure_getNumberOfTrianglesInTriangle() == element_data_stack_out->getNumberOfElementsOnStack());

		// clear forward stack and SWAP to make the output stack the forward stack
		cStacks->element_data_stacks.forward.clear();
		cStacks->element_data_stacks.swap();

#if ADAPTIVE_SUBPARTITION_STACKS
		/*
		 * if this assertion fails, add some offset to estimated_num_triangles
		 */
		assert (estimated_num_triangles == cStacks->element_data_stacks.forward.getNumberOfElementsOnStack());

		cStacks->changeStackSize_AfterAdaptivityLastTraversal();
#endif

		return maxLeafTraversalDepth;
	}
};



/**
 * adaptive refinement
 */
template <typename CKernelClass, typename CSimulationStacks>
class CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData
{
public:
	CKernelClass kernelClass;
	typedef typename CSimulationStacks::TElementData TElementData;

	CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_FirstTraversal<CKernelClass, CSimulationStacks> firstTraversal;
	CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_MiddleTraversals<CKernelClass, CSimulationStacks> middleTraversals;
	CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData_LastTraversal<CKernelClass, CSimulationStacks> lastTraversal;


	CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData()	:
		firstTraversal(kernelClass),
		middleTraversals(kernelClass),
		lastTraversal(kernelClass)
	{
	}



	/**
	 * setup the initial partition traversal for the given factory
	 */
	void setup_sfcMethods(
			CTriangle_Factory &i_triangleFactory
	)
	{
		firstTraversal.setup_sfcMethods(i_triangleFactory);
		middleTraversals.setup_sfcMethods(i_triangleFactory);
		lastTraversal.setup_sfcMethods(i_triangleFactory);
	}


	/**
	 * setup the limiters for minimum and maximum depth of curves
	 */
	void setup_RootTraversator(
			int i_depth_limiter_min,
			int i_depth_limiter_max
	)
	{
		firstTraversal.setup_TraversatorParameters(i_depth_limiter_min, i_depth_limiter_max);
	}


	/**
	 * setup the initial partition traversal for the given factory
	 */
	void setup_Partition(
			CAdaptiveConformingGrid_Vertices_DepthLimiter_ElementData<CKernelClass, CSimulationStacks> &i_parent,
			CTriangle_Factory &p_triangleFactory
	)
	{
		firstTraversal.setup_Partition(i_parent.firstTraversal, p_triangleFactory);
		middleTraversals.setup_Partition(i_parent.middleTraversals, p_triangleFactory);
		lastTraversal.setup_Partition(i_parent.lastTraversal, p_triangleFactory);

		// setup kernel class
		kernelClass.setup_WithKernel(i_parent.kernelClass);

		firstTraversal.setup_TraversatorParameters(i_parent.firstTraversal.depth_limiter_min, i_parent.firstTraversal.depth_limiter_max);
	}
};

}
}

#endif
