/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpi project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpi
 *
 *
 *
 *  Created on: Nov 20, 2011
 *      Author: kleinfl
 */

#ifndef CSIMULATIONTSUNAMI_SERIAL_REGULAR_TRAVERSE_HPP_
#define CSIMULATIONTSUNAMI_SERIAL_REGULAR_TRAVERSE_HPP_


#include "../tsunami_common/types/CTsunamiTypes.hpp"
#include "../../libsierpi/triangle/CTriangle_Normals.hpp"


#include "../CSimulationCommon.hpp"

#include "../tsunami_common/kernels/simulation/CEdgeComm_Tsunami.hpp"


#include "CSimulationTsunami_Serial_Regular_Traverse_Table.hpp"


namespace sierpi
{
namespace travs
{
// scons --cppcompiler=gnu --threading=off --gui=on --mode=debug --simulation=tsunami_serial_regular -j8
// scons --cppcompiler=gcc --threading=off --gui=off --mode=release --simulation=tsunami_serial_regular -j8

/**
 * edge type (data found on...): left stack, right stack, border or inner domain triangle
 */
enum edge_type {LEFT, RIGHT, BORDER, INNER};


class CSimulationTsunami_Serial_Regular_Traverse
{

	/**
	 * \brief setup parameters for simulation
	 */
public:
	inline void setup_WithParameters(
			TTsunamiDataScalar i_timestep_size,			///< timestep size
			TTsunamiDataScalar i_square_side_length,	///< length of a square (a single catheti)
			TTsunamiDataScalar i_gravity = 9.81			///< gravitation constant
	)
	{
		cEdgeComm_Tsunami.setup_WithParameters(i_timestep_size, i_square_side_length, i_gravity);
		cTsunami_EdgeComm_Serial_Regular_Traverse_Table.setup_WithParameters(i_timestep_size, i_square_side_length, i_gravity);
	}

	/**
	 * \brief setup kernel boundary condition
	 */
public:
	inline void setup_WithParameters_KernelBoundaryDirichlet(
			const CTsunamiEdgeData *p_boundary_dirichlet
	)
	{
		cEdgeComm_Tsunami.setup_WithParameters_KernelBoundaryDirichlet(p_boundary_dirichlet);
		cTsunami_EdgeComm_Serial_Regular_Traverse_Table.setup_WithParameters_KernelBoundaryDirichlet(p_boundary_dirichlet);
	}

public:
	CSimulationStacks<CTsunamiElementData,CTsunamiEdgeData> *cStacks;
	sierpi::kernels::CEdgeComm_Tsunami cEdgeComm_Tsunami;
	sierpi::travs::CSimulationTsunami_Serial_Regular_Traverse_Table cTsunami_EdgeComm_Serial_Regular_Traverse_Table;
	int recursionDepth;


	/**
	 * calculates domain size for given depth
	 */
	inline int size(int depth){
		return 1 << depth;
	}

	/**
	 * determines if the searched index lies in the first half of the current triangle
	 */
	inline bool first_half(int index, int size, int offset){
		return (index - offset < (size >> 1));
	}

	/**
	 * determines if triangle has odd refinement depth
	 */
	inline bool odd_depth(int depth){
		return (depth & 1) == 1;
	}

	/**
	 * calculates normal direction of refined triangle
	 */
	inline int calc_normal(int normal, bool odd, bool first_half){
		if (!((first_half && !odd) || (!first_half && odd)))
			return (normal + 3) & 7;
		return (normal + 5) & 7;
	}

	/**
	 * calculates index gap on stacks for second partition of the refined triangle
	 */
	inline int calc_stack_gap(int depth){
		if (odd_depth(depth))
			return size((depth - 1) >> 1);
		return size((depth >> 1) - 1);
	}

	/**
	 * determines mid element of left stack
	 * used only for initial starting indices
	 */
	inline int get_mid_left(int depth){
		if (odd_depth(depth))
			return size((depth - 1) >> 1);
		return size(depth >> 1);
	}

	/**
	 * recursively determines neighbor indices for a given triangle search index
	 * starting with the whole search domain, partitioning to half size in every step
	 */
	void triangle_recursion_H_EVEN(
			edge_type edge_type_hyp,
			edge_type edge_type_right,
			edge_type edge_type_left,
			int index_hyp,
			int index_right,
			int index_left,
			int triangle_size,
			int index_local,
			int normal,
			int depth
	){
		/**
		 * further refinement possible: descend into half-size partition
		 */
		if (triangle_size > 1){
			// new variables for next recursion step
			int index_hyp_new, index_right_new, index_left_new, normal_new;
			int triangle_size_new = triangle_size >> 1;
			int index_local_new = index_local;

			// determine new normal
			normal_new = calc_normal(normal, false, true);

			/**
			 * determine new triangle type, edge types and starting indices of neighbor partitions
			 * depending on current triangle type, following the mapping
			 * H -> (V',K'); K -> (H',V'); V -> (H',K'); H' -> (V,K); K' -> (H,V); V' -> (H,K)
			 */
			switch (edge_type_hyp){
				case INNER: index_left_new = index_hyp + triangle_size_new; break;
				case RIGHT: index_left_new = index_hyp + calc_stack_gap(depth); break;
				case BORDER: index_left_new = 0; break;
				default:	break;
			}
			switch (edge_type_right){
				case INNER: index_hyp_new = index_right + triangle_size_new; break;
				case LEFT: index_hyp_new = index_right; break;
				case RIGHT: index_hyp_new = index_left; break;
				case BORDER: index_hyp_new = 0; break;
				default:	break;
			}
			index_right_new = index_local + triangle_size_new;

			// recursion call with new data
			triangle_recursion_K_ODD(edge_type_right, INNER, edge_type_hyp,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);


			// determine new normal
			normal_new = calc_normal(normal, false, false);

			switch (edge_type_hyp){
				case INNER: index_right_new = index_hyp; break;
				case RIGHT: index_right_new = index_hyp; break;
				case BORDER: index_right_new = 0; break;
				default:	break;
			}
			switch (edge_type_left){
				case INNER: index_hyp_new = index_left; break;
				case LEFT: index_hyp_new = index_left; break;
				case RIGHT: index_hyp_new = index_right; break;
				case BORDER: index_hyp_new = 0; break;
				default:	break;
			}
			index_left_new = index_local;

			// adjust local index
			index_local_new = index_local + triangle_size_new;

			// recursion call with new data
			triangle_recursion_V_ODD(edge_type_left, edge_type_hyp, INNER,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);

		}
		/**
		 * recursion termination: size 1, no further refinement possible
		 * searched neighbor indices = start indices of neighbor partitions
		 */
		else{
			call_elementAction(edge_type_hyp, edge_type_right, edge_type_left, index_hyp, index_right, index_left, index_local, normal);
		}
	}


	void triangle_recursion_K_EVEN(
			edge_type edge_type_hyp,
			edge_type edge_type_right,
			edge_type edge_type_left,
			int index_hyp,
			int index_right,
			int index_left,
			int triangle_size,
			int index_local,
			int normal,
			int depth
	){
		/**
		 * further refinement possible: descend into half-size partition
		 */
		if (triangle_size > 1){
			// new variables for next recursion step
			int index_hyp_new, index_right_new, index_left_new, normal_new;
			int triangle_size_new = triangle_size >> 1;
			int index_local_new = index_local;

			// determine new normal
			normal_new = calc_normal(normal, false, true);

			/**
			 * determine new triangle type, edge types and starting indices of neighbor partitions
			 * depending on current triangle type, following the mapping
			 * H -> (V',K'); K -> (H',V'); V -> (H',K'); H' -> (V,K); K' -> (H,V); V' -> (H,K)
			 */
			switch (edge_type_hyp){
				case INNER: index_left_new = index_hyp + triangle_size_new; break;
				case RIGHT: index_left_new = index_hyp + calc_stack_gap(depth); break;
				case BORDER: index_left_new = 0; break;
				default:	break;
			}
			switch (edge_type_right){
				case INNER: index_hyp_new = index_right + triangle_size_new; break;
				case LEFT: index_hyp_new = index_right; break;
				case RIGHT: index_hyp_new = index_left; break;
				case BORDER: index_hyp_new = 0; break;
				default:	break;
			}
			index_right_new = index_local + triangle_size_new;

			// recursion call with new data
			triangle_recursion_V_ODD(edge_type_right, INNER, edge_type_hyp,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);


			// determine new normal
			normal_new = calc_normal(normal, false, false);

			switch (edge_type_hyp){
				case INNER: index_right_new = index_hyp; break;
				case RIGHT: index_right_new = index_hyp; break;
				case BORDER: index_right_new = 0; break;
				default:	break;
			}
			switch (edge_type_left){
				case INNER: index_hyp_new = index_left; break;
				case LEFT: index_hyp_new = index_left; break;
				case RIGHT: index_hyp_new = index_right; break;
				case BORDER: index_hyp_new = 0; break;
				default:	break;
			}
			index_left_new = index_local;

			// adjust local index
			index_local_new = index_local + triangle_size_new;

			// recursion call with new data
			triangle_recursion_H_ODD(edge_type_left, edge_type_hyp, INNER,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);

		}
		/**
		 * recursion termination: size 1, no further refinement possible
		 * searched neighbor indices = start indices of neighbor partitions
		 */
		else{
			call_elementAction(edge_type_hyp, edge_type_right, edge_type_left, index_hyp, index_right, index_left, index_local, normal);
		}
	}


	void triangle_recursion_V_EVEN(
			edge_type edge_type_hyp,
			edge_type edge_type_right,
			edge_type edge_type_left,
			int index_hyp,
			int index_right,
			int index_left,
			int triangle_size,
			int index_local,
			int normal,
			int depth
	){
		/**
		 * further refinement possible: descend into half-size partition
		 */
		if (triangle_size > 1){
			// new variables for next recursion step
			int index_hyp_new, index_right_new, index_left_new, normal_new;
			int triangle_size_new = triangle_size >> 1;
			int index_local_new = index_local;

			// determine new normal
			normal_new = calc_normal(normal, false, true);

			/**
			 * determine new triangle type, edge types and starting indices of neighbor partitions
			 * depending on current triangle type, following the mapping
			 * H -> (V',K'); K -> (H',V'); V -> (H',K'); H' -> (V,K); K' -> (H,V); V' -> (H,K)
			 */
			switch (edge_type_hyp){
				case INNER: index_left_new = index_hyp + triangle_size_new; break;
				case RIGHT: index_left_new = index_hyp + calc_stack_gap(depth); break;
				case BORDER: index_left_new = 0; break;
				default:	break;
			}
			switch (edge_type_right){
				case INNER: index_hyp_new = index_right + triangle_size_new; break;
				case LEFT: index_hyp_new = index_right; break;
				case RIGHT: index_hyp_new = index_left; break;
				case BORDER: index_hyp_new = 0; break;
				default:	break;
			}
			index_right_new = index_local + triangle_size_new;

			// recursion call with new data
			triangle_recursion_K_ODD(edge_type_right, INNER, edge_type_hyp,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);


			// determine new normal
			normal_new = calc_normal(normal, false, false);

			switch (edge_type_hyp){
				case INNER: index_right_new = index_hyp; break;
				case RIGHT: index_right_new = index_hyp; break;
				case BORDER: index_right_new = 0; break;
				default:	break;
			}
			switch (edge_type_left){
				case INNER: index_hyp_new = index_left; break;
				case LEFT: index_hyp_new = index_left; break;
				case RIGHT: index_hyp_new = index_right; break;
				case BORDER: index_hyp_new = 0; break;
				default:	break;
			}
			index_left_new = index_local;

			// adjust local index
			index_local_new = index_local + triangle_size_new;

			// recursion call with new data
			triangle_recursion_H_ODD(edge_type_left, edge_type_hyp, INNER,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);

		}
		/**
		 * recursion termination: size 1, no further refinement possible
		 * searched neighbor indices = start indices of neighbor partitions
		 */
		else{
			call_elementAction(edge_type_hyp, edge_type_right, edge_type_left, index_hyp, index_right, index_left, index_local, normal);
		}
	}


	void triangle_recursion_H_ODD(
			edge_type edge_type_hyp,
			edge_type edge_type_right,
			edge_type edge_type_left,
			int index_hyp,
			int index_right,
			int index_left,
			int triangle_size,
			int index_local,
			int normal,
			int depth
	){
		/**
		 * further refinement possible: descend into half-size partition
		 */
		if (triangle_size > 1){
			// new variables for next recursion step
			int index_hyp_new, index_right_new, index_left_new, normal_new;
			int triangle_size_new = triangle_size >> 1;
			int index_local_new = index_local;

			// determine new normal
			normal_new = calc_normal(normal, true, true);

			/**
			 * determine new triangle type, edge types and starting indices of neighbor partitions
			 * depending on current triangle type, following the mapping
			 * H -> (V',K'); K -> (H',V'); V -> (H',K'); H' -> (V,K); K' -> (H,V); V' -> (H,K)
			 */
			switch (edge_type_hyp){
				case INNER: index_right_new = index_hyp + triangle_size_new; break;
				case LEFT: index_right_new = index_hyp + calc_stack_gap(depth); break;
				case BORDER: index_right_new = 0; break;
				default:	break;
			}
			switch (edge_type_left){
				case INNER: index_hyp_new = index_left + triangle_size_new; break;
				case LEFT: index_hyp_new = index_right; break;
				case RIGHT: index_hyp_new = index_left; break;
				case BORDER: index_hyp_new = 0; break;
				default:	break;
			}
			index_left_new = index_local + triangle_size_new;

			// recursion call with new data
			triangle_recursion_K_EVEN(edge_type_left, edge_type_hyp, INNER,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);

			// determine new normal
			normal_new = calc_normal(normal, true, false);

			// adjust local index
			index_local_new = index_local + triangle_size_new;

			switch (edge_type_hyp){
				case INNER: index_left_new = index_hyp; break;
				case LEFT: index_left_new = index_hyp; break;
				case BORDER: index_left_new = 0; break;
				default:	break;
			}
			switch (edge_type_right){
				case INNER: index_hyp_new = index_right; break;
				case LEFT: index_hyp_new = index_left; break;
				case RIGHT: index_hyp_new = index_right; break;
				case BORDER: index_hyp_new = 0; break;
			}
			index_right_new = index_local;

			// recursion call with new data
			triangle_recursion_V_EVEN(edge_type_right, INNER, edge_type_hyp,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);

		}
		/**
		 * recursion termination: size 1, no further refinement possible
		 * searched neighbor indices = start indices of neighbor partitions
		 */
		else{
			call_elementAction(edge_type_hyp, edge_type_right, edge_type_left, index_hyp, index_right, index_left, index_local, normal);
		}
	}


	void triangle_recursion_K_ODD(
			edge_type edge_type_hyp,
			edge_type edge_type_right,
			edge_type edge_type_left,
			int index_hyp,
			int index_right,
			int index_left,
			int triangle_size,
			int index_local,
			int normal,
			int depth
	){
		/**
		 * further refinement possible: descend into half-size partition
		 */
		if (triangle_size > 1){
			// new variables for next recursion step
			int index_hyp_new, index_right_new, index_left_new, normal_new;
			int triangle_size_new = triangle_size >> 1;
			int index_local_new = index_local;

			// determine new normal
			normal_new = calc_normal(normal, true, true);

			/**
			 * determine new triangle type, edge types and starting indices of neighbor partitions
			 * depending on current triangle type, following the mapping
			 * H -> (V',K'); K -> (H',V'); V -> (H',K'); H' -> (V,K); K' -> (H,V); V' -> (H,K)
			 */
			switch (edge_type_hyp){
				case INNER: index_right_new = index_hyp + triangle_size_new; break;
				case LEFT: index_right_new = index_hyp + calc_stack_gap(depth); break;
				case BORDER: index_right_new = 0; break;
				default:	break;
			}
			switch (edge_type_left){
				case INNER: index_hyp_new = index_left + triangle_size_new; break;
				case LEFT: index_hyp_new = index_right; break;
				case RIGHT: index_hyp_new = index_left; break;
				case BORDER: index_hyp_new = 0; break;
				default:	break;
			}
			index_left_new = index_local + triangle_size_new;

			// recursion call with new data
			triangle_recursion_V_EVEN(edge_type_left, edge_type_hyp, INNER,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);

			// determine new normal
			normal_new = calc_normal(normal, true, false);

			// adjust local index
			index_local_new = index_local + triangle_size_new;

			switch (edge_type_hyp){
				case INNER: index_left_new = index_hyp; break;
				case LEFT: index_left_new = index_hyp; break;
				case BORDER: index_left_new = 0; break;
				default:	break;
			}
			switch (edge_type_right){
				case INNER: index_hyp_new = index_right; break;
				case LEFT: index_hyp_new = index_left; break;
				case RIGHT: index_hyp_new = index_right; break;
				case BORDER: index_hyp_new = 0; break;
			}
			index_right_new = index_local;

			// recursion call with new data
			triangle_recursion_H_EVEN(edge_type_right, INNER, edge_type_hyp,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);

		}
		/**
		 * recursion termination: size 1, no further refinement possible
		 * searched neighbor indices = start indices of neighbor partitions
		 */
		else{
			call_elementAction(edge_type_hyp, edge_type_right, edge_type_left, index_hyp, index_right, index_left, index_local, normal);
		}
	}


	void triangle_recursion_V_ODD(
			edge_type edge_type_hyp,
			edge_type edge_type_right,
			edge_type edge_type_left,
			int index_hyp,
			int index_right,
			int index_left,
			int triangle_size,
			int index_local,
			int normal,
			int depth
	){
		/**
		 * further refinement possible: descend into half-size partition
		 */
		if (triangle_size > 1){
			// new variables for next recursion step
			int index_hyp_new, index_right_new, index_left_new, normal_new;
			int triangle_size_new = triangle_size >> 1;
			int index_local_new = index_local;

			// determine new normal
			normal_new = calc_normal(normal, true, true);

			/**
			 * determine new triangle type, edge types and starting indices of neighbor partitions
			 * depending on current triangle type, following the mapping
			 * H -> (V',K'); K -> (H',V'); V -> (H',K'); H' -> (V,K); K' -> (H,V); V' -> (H,K)
			 */
			switch (edge_type_hyp){
				case INNER: index_right_new = index_hyp + triangle_size_new; break;
				case LEFT: index_right_new = index_hyp + calc_stack_gap(depth); break;
				case BORDER: index_right_new = 0; break;
				default:	break;
			}
			switch (edge_type_left){
				case INNER: index_hyp_new = index_left + triangle_size_new; break;
				case LEFT: index_hyp_new = index_right; break;
				case RIGHT: index_hyp_new = index_left; break;
				case BORDER: index_hyp_new = 0; break;
				default:	break;
			}
			index_left_new = index_local + triangle_size_new;

			// recursion call with new data
			triangle_recursion_K_EVEN(edge_type_left, edge_type_hyp, INNER,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);

			// determine new normal
			normal_new = calc_normal(normal, true, false);

			// adjust local index
			index_local_new = index_local + triangle_size_new;

			switch (edge_type_hyp){
				case INNER: index_left_new = index_hyp; break;
				case LEFT: index_left_new = index_hyp; break;
				case BORDER: index_left_new = 0; break;
				default:	break;
			}
			switch (edge_type_right){
				case INNER: index_hyp_new = index_right; break;
				case LEFT: index_hyp_new = index_left; break;
				case RIGHT: index_hyp_new = index_right; break;
				case BORDER: index_hyp_new = 0; break;
			}
			index_right_new = index_local;

			// recursion call with new data
			triangle_recursion_H_EVEN(edge_type_right, INNER, edge_type_hyp,
					index_hyp_new, index_right_new, index_left_new,
					triangle_size_new, index_local_new, normal_new, depth-1);

		}
		/**
		 * recursion termination: size 1, no further refinement possible
		 * searched neighbor indices = start indices of neighbor partitions
		 */
		else{
			call_elementAction(edge_type_hyp, edge_type_right, edge_type_left, index_hyp, index_right, index_left, index_local, normal);
		}
	}



	/**
	 * prepares resulting data and calls corresponding elementAction_XXX method
	 */
	inline void call_elementAction(
			edge_type edge_type_hyp,
			edge_type edge_type_right,
			edge_type edge_type_left,
			int index_hyp,
			int index_right,
			int index_left,
			int search_index,
			int normal_hyp
	){
		// invert indices to switch root triangle entrance from left to right
		/*int maxNum = cStacks->element_data_stacks.forward.getNumberOfElementsOnStack() - 1;
		index_hyp = maxNum - index_hyp;
		index_right = maxNum - index_right;
		index_left = maxNum - index_left;
		search_index = maxNum - search_index;*/

		// calc normal direction for left/right edge
		int normal_left = (normal_hyp + 3) & 7;
		int normal_right = (normal_hyp + 5) & 7;

		// get data of current element from forward stack and store (copy) it to backward stack
		//CTsunamiElementData current_element = &(cStacks->element_data_stacks.forward[search_index]);
		CTsunamiElementData *current_element = &(cStacks->element_data_stacks.backward[search_index]);
		cStacks->element_data_stacks.forward.getElementAtIndex(search_index, *current_element);

		// edges on border?
		bool hyp_border = edge_type_hyp != INNER;
		bool right_border = edge_type_right != INNER;
		bool left_border = edge_type_left != INNER;

		// combine bool values to flag-int for switch statement
		int border_flags = (((int) hyp_border) << 2) + (((int) right_border) << 1) + ((int) left_border);

		// get edge data of neighbor elements from forward stack
		CTsunamiEdgeData edge_hyp;
		CTsunamiEdgeData edge_right;
		CTsunamiEdgeData edge_left;

		if (!hyp_border)
			cEdgeComm_Tsunami.storeHypCommData(
					CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
					CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
					CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
					recursionDepth,
					&(cStacks->element_data_stacks.forward[index_hyp]),
					&edge_hyp
				);

		if (!right_border)
			cEdgeComm_Tsunami.storeLeftEdgeCommData(
					CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
					CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
					CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
					recursionDepth,
					&(cStacks->element_data_stacks.forward[index_right]),
					&edge_right
				);

		if (!left_border)
			cEdgeComm_Tsunami.storeRightEdgeCommData(
					CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
					CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
					CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
					recursionDepth,
					&(cStacks->element_data_stacks.forward[index_left]),
					&edge_left
				);

		#if 1
			#if COMPILE_WITH_DEBUG_TSUNAMI_ELEMENTDATA_VALIDATION
				//std::cout << "REC INFO (" << index_hyp << ", " << index_right << ", " << index_left << ") " << search_index << std::endl;
			/*validate_normals(
					CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
					CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
					CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
					current_element);*/
#if 0
				std::cout << "NORMAL DIRECTIONS (H,R,L): (" << normal_hyp << ", " << normal_right << ", " << normal_left << "), INDEX: " << search_index << std::endl;
				std::cout << "NEIGHBORS (H,R,L): (" << index_hyp << ", " << index_right << ", " << index_left << ")" << std::endl;
				//std::cout << "vertices: ";
				//for(int i=0; i<=2; i++){
				//	std::cout << "(" << current_element->vertices[i][0] << ", " << current_element->vertices[i][1] << ") ";
				//} std::cout << std::endl;//std::cout << "flags: " << border_flags << " " << maxNum << std::endl;
#endif
				//std::cout << "INFO " << search_index << ": h " << current_element->hyp_edge.qy << ", l " << current_element->left_edge.qy << ", r " << current_element->right_edge.qx << std::endl;
			#endif
		#endif

		// call elementAction_XXX depending on edge types
		switch (border_flags){
			case 0:
				cEdgeComm_Tsunami.elementAction_EEE(
						CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
						CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
						CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
						recursionDepth,
						current_element,
						&edge_hyp, &edge_right, &edge_left
					);
				break;

			case 1:
				cEdgeComm_Tsunami.elementAction_EEB(
						CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
						CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
						CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
						recursionDepth,
						current_element,
						&edge_hyp, &edge_right
					);
				break;

			case 2:
				cEdgeComm_Tsunami.elementAction_EBE(
						CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
						CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
						CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
						recursionDepth,
						current_element,
						&edge_hyp,
						&edge_left
					);
				break;

			case 3:
				cEdgeComm_Tsunami.elementAction_EBB(
						CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
						CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
						CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
						recursionDepth, current_element, &edge_hyp);
				break;
			case 4:
				cEdgeComm_Tsunami.elementAction_BEE(
						CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
						CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
						CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
						recursionDepth, current_element, &edge_right, &edge_left);
				break;
			case 5:
				cEdgeComm_Tsunami.elementAction_BEB(
						CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
						CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
						CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
						recursionDepth, current_element, &edge_right);
				break;
			case 6:
				cEdgeComm_Tsunami.elementAction_BBE(
						CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
						CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
						CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
						recursionDepth, current_element, &edge_left);
				break;
			case 7:
				cEdgeComm_Tsunami.elementAction_BBB(
						CTriangle_Normals::normal_table[normal_hyp][0], CTriangle_Normals::normal_table[normal_hyp][1],
						CTriangle_Normals::normal_table[normal_right][0], CTriangle_Normals::normal_table[normal_right][1],
						CTriangle_Normals::normal_table[normal_left][0], CTriangle_Normals::normal_table[normal_left][1],
						recursionDepth, current_element);
				break;
		}
	}

/*	inline void validate_normals(float hyp_x, float hyp_y, float right_x, float right_y, float left_x, float left_y, CTsunamiElementData *element){
#if COMPILE_WITH_DEBUG_TSUNAMI_ELEMENTDATA_VALIDATION
		float normal_coord [] = {hyp_x, hyp_y, right_x, right_y, left_x, left_y};

		float nrx = element->vertices[2][0] - element->vertices[0][0];
		float nry = element->vertices[2][1] - element->vertices[0][1];
		float lr = CMath::sqrt(nrx*nrx + nry*nry);
		nrx = nrx / lr;
		nry = nry / lr;
		float nlx = element->vertices[2][0] - element->vertices[1][0];
		float nly = element->vertices[2][1] - element->vertices[1][1];
		float ll = CMath::sqrt(nlx*nlx + nly*nly);
		nlx = nlx / ll;
		nly = nly / ll;
		float midhx = element->vertices[0][0] + (element->vertices[1][0] - element->vertices[0][0]) / 2;
		float midhy = element->vertices[0][1] + (element->vertices[1][1] - element->vertices[0][1]) / 2;
		float nhx = midhx - element->vertices[2][0];
		float nhy = midhy - element->vertices[2][1];
		float lh = CMath::sqrt(nhx*nhx + nhy*nhy);
		nhx = nhx / lh;
		nhy = nhy / lh;

		float eps = 0.0001;
		bool assert_failed = false;
		if (CMath::abs(nhx - normal_coord[0]) > eps)
		{
			std::cout << "NORMAL DIFF on HYPx" << std::endl;
			assert_failed = true;
		}
		if (CMath::abs(nhy - normal_coord[1]) > eps)
		{
			std::cout << "NORMAL DIFF on HYPy" << std::endl;
			assert_failed = true;
		}

		if (CMath::abs(nrx - normal_coord[2]) > eps)
		{
			std::cout << "NORMAL DIFF on RIGHTx" << std::endl;
			assert_failed = true;
		}
		if (CMath::abs(nry - normal_coord[3]) > eps)
		{
			std::cout << "NORMAL DIFF on RIGHTy" << std::endl;
			assert_failed = true;
		}

		if (CMath::abs(nlx - normal_coord[4]) > eps)
		{
			std::cout << "NORMAL DIFF on LEFTx" << std::endl;
			assert_failed = true;
		}
		if (CMath::abs(nly - normal_coord[5]) > eps)
		{
			std::cout << "NORMAL DIFF on LEFTy" << std::endl;
			assert_failed = true;
		}

		if (!assert_failed)
			return;

		std::cout << nhx << " " << nhy << std::endl;
		std::cout << "NORMAL H: arr (" << normal_coord[0] << ", " << normal_coord[1] << ") ,calc (" << nhx << ", " << nhy << ")" << std::endl;
		std::cout << nrx << " " << nry << std::endl;
		std::cout << "NORMAL R: arr (" << normal_coord[2] << ", " << normal_coord[3] << ") ,calc (" << nrx << ", " << nry << ")" << std::endl;
		std::cout << nlx << " " << nly << std::endl;
		std::cout << "NORMAL L: arr (" << normal_coord[4] << ", " << normal_coord[5] << ") ,calc (" << nlx << ", " << nly << ")" << std::endl;
#endif
	}*/

	/**
	 * traverses domain with given depth and element data stacks
	 * exchanges edge data
	 */
	inline void traverse(
			CSimulationStacks<CTsunamiElementData,CTsunamiEdgeData> *i_cSimulationStacks,
			int depth,
			edge_type edge_type_hyp,
			edge_type edge_type_right,
			edge_type edge_type_left
	){
		cStacks = i_cSimulationStacks;
		recursionDepth = depth;
		int triangle_size = size(depth);


		// validate number of elements on stack
		assert(i_cSimulationStacks->element_data_stacks.getNumberOfElementsOnStack() == (1 << depth));

		// starting conditions: whole domain (given edge types with pseudo indices), domain size, local index 0, normal direction 1
		triangle_recursion_K_EVEN(edge_type_hyp, edge_type_right, edge_type_left, 0, 0, 0, triangle_size, 0, 1, depth);

		// set stack element counters
		cStacks->element_data_stacks.backward.setStackElementCounter(cStacks->element_data_stacks.forward.getNumberOfElementsOnStack());
		cStacks->element_data_stacks.forward.setStackElementCounter(0);

		// swap forward/backward stacks for next traversal
		cStacks->element_data_stacks.swap();
	}


	inline void traverse(
			CSimulationStacks<CTsunamiElementData,CTsunamiEdgeData> *i_cSimulationStacks,
			int depth
	){
		// simple case: whole simulation domain (3x border)
		traverse(i_cSimulationStacks, depth, BORDER, BORDER, BORDER);
	}

	/**
	 * traverse using predefined arrays
	 */
	inline void traverse_withList(CSimulationStacks<CTsunamiElementData,CTsunamiEdgeData> *i_cSimulationStacks, int depth){

		cTsunami_EdgeComm_Serial_Regular_Traverse_Table.traverse(i_cSimulationStacks, depth);

	}


};

}
}

#endif /* CTSUNAMI_HPP_ */

